from scfile.consts import Magic
from scfile.enums import StructFormat as F
from scfile.file.data import TextureData
from scfile.file.encoder import FileEncoder

from .structure import DDS


RGBA8 = [0xFF, 0xFF00, 0xFF0000, 0xFF]


class DdsEncoder(FileEncoder[TextureData]):
    @property
    def magic(self):
        return Magic.DDS

    def serialize(self):
        self.b.writeb(F.U32, DDS.HEADER.SIZE)
        self.b.writeb(F.U32, self.flags)
        self.b.writeb(F.U32, self.data.height)
        self.b.writeb(F.U32, self.data.width)
        self.b.writeb(F.U32, self.pitch_or_linear_size)

        self.b.writen(1)  # Depth
        self.b.writeb(F.U32, 1)  # MipMapsCount
        self.b.writen(11)  # Reserved

        self.b.writeb(F.U32, DDS.PF.SIZE)

        if self.compressed:
            self.b.writeb(F.U32, DDS.PF.FLAG.FOURCC)
            self.b.write(self.data.fourcc)  # FourCC
            self.b.writen(5)  # BitCount & BitMasks

        else:
            self.b.writeb(F.U32, DDS.PF.RGB_FLAGS)
            self.b.writen(1)  # FourCC
            self.b.writeb(F.U32, DDS.PF.BIT_COUNT)  # BitCount

            # BitMasks
            for mask in RGBA8:
                self.b.writeb(F.U32, mask)

        self.b.writeb(F.U32, DDS.TEXTURE | DDS.COMPLEX | DDS.MIPMAP)  # Caps1
        self.b.writeb(F.U32, self.cubemap)  # Caps2
        self.b.writen(3)  # Reserved

        self.b.write(self.data.image)

    @property
    def compressed(self) -> bool:
        return b"DXT" in self.data.fourcc

    @property
    def flags(self) -> int:
        if self.compressed:
            return DDS.HEADER.FLAGS | DDS.HEADER.FLAG.LINEARSIZE
        return DDS.HEADER.FLAGS | DDS.HEADER.FLAG.PITCH

    @property
    def pitch(self) -> int:
        bytes_per_pixel = 4
        aligned_width = (self.data.width * bytes_per_pixel + 3) & ~3
        return aligned_width

    @property
    def pitch_or_linear_size(self) -> int:
        if self.compressed:
            return self.data.linear_size
        return self.pitch

    @property
    def cubemap(self) -> int:
        return 0



from typing import Any

import lz4.block  # type: ignore
from quicktex import RawTexture  # type: ignore
from quicktex.s3tc.bc3 import BC3Encoder  # type: ignore
from quicktex.s3tc.bc5 import BC5Decoder, BC5Texture  # type: ignore

from scfile import exceptions as exc
from scfile.consts import Signature
from scfile.enums import ByteOrder
from scfile.enums import StructFormat as F
from scfile.file.data import TextureData
from scfile.file.dds.encoder import DdsEncoder
from scfile.file.decoder import FileDecoder
from scfile.io.ol import OlFileIO

from .converter.base import RGBA8Converter
from .converter.bgra8 import BGRA8Converter
from .converter.rgba32f import RGBA32FConverter
from .formats import SUPPORTED_FORMATS


class OlDecoder(FileDecoder[OlFileIO, TextureData]):
    CONVERT_TO_RGBA8 = True
    CONVERT_TO_DXT5 = False

    def to_dds(self):
        return self.convert_to(DdsEncoder)

    @property
    def opener(self):
        return OlFileIO

    @property
    def order(self):
        return ByteOrder.BIG

    @property
    def signature(self):
        return Signature.OL

    def create_data(self):
        return TextureData(self.width, self.height, self.linear_size, self.fourcc, self.image)

    def parse(self):
        self._parse_header()
        self._parse_fourcc()
        self._parse_sizes()
        self._parse_image()
        self._convert()

    def _parse_header(self):
        self.width = self.f.readb(F.U32)
        self.height = self.f.readb(F.U32)
        self.mipmap_count = self.f.readb(F.U32)

    def _parse_fourcc(self):
        self.fourcc = self.f.readfourcc()

        if self.fourcc not in SUPPORTED_FORMATS:
            raise exc.OlUnknownFourcc(self.path, self.fourcc.decode())

    def _read_sizes(self):
        return [self.f.readb(F.U32) for _ in range(self.mipmap_count)]

    def _parse_sizes(self):
        self.uncompressed = self._read_sizes()
        self.compressed = self._read_sizes()

    def _decompress_image(self):
        compressed = self.f.read(self.compressed[0])
        uncompressed_size = self.uncompressed[0]

        self.image = bytes(lz4.block.decompress(compressed, uncompressed_size))  # type: ignore

    def _parse_image(self):
        try:
            self.texture_id = self.f.reads()
            self._decompress_image()

        except Exception:
            raise exc.OlInvalidFormat(self.path)

    def _convert(self):
        if self.CONVERT_TO_RGBA8:
            self._to_rgba8()

        if self.CONVERT_TO_DXT5:
            self._to_dxt5()

    @property
    def linear_size(self) -> int:
        return self.uncompressed[0]

    @property
    def is_compressed(self) -> bool:
        return b"DXT" in self.fourcc

    @property
    def imagedata(self) -> tuple[bytes, int, int]:
        return (self.image, self.width, self.height)

    @property
    def bc5texture(self) -> Any:
        return BC5Texture.from_bytes(*self.imagedata)  # type: ignore

    @property
    def rawtexture(self) -> Any:
        return RawTexture.frombytes(*self.imagedata)  # type: ignore

    def _to_rgba8(self):
        match self.fourcc:
            case b"BGRA8":
                self.image = BGRA8Converter(*self.imagedata).to_rgba8()

            case b"RGBA32F":
                self.image = RGBA32FConverter(*self.imagedata).to_rgba8()

            case b"DXN_XY":
                self.image = bytes(BC5Decoder().decode(self.bc5texture))  # type: ignore
                self.image = RGBA8Converter(*self.imagedata).invert()

            case _:
                pass

    def _to_dxt5(self):
        if not self.is_compressed:
            self.fourcc = b"DXT5"
            self.image = bytes(BC3Encoder(level=9).encode(self.rawtexture))  # type: ignore



import numpy as np

from .base import RGBA8Converter


class RGBA32FConverter(RGBA8Converter):
    dtype = np.float32

    def to_rgba8(self) -> bytes:
        rgba = (self.array * 255).astype(np.uint8)
        return rgba.tobytes()



from scfile.consts import McsaModel
from scfile.enums import StructFormat as F
from scfile.exceptions.mcsa import McsaCountsLimit
from scfile.io.binary import BinaryFileIO


class McsaFileIO(BinaryFileIO):
    def readvalues(self, fmt: str, size: int, count: int):
        """Read repetitive data values."""
        return self.unpack(f"{size*count}{fmt}")

    def readcounts(self):
        """Read u32 value and validates count limit."""
        counts = self.readb(F.U32)
        if counts > McsaModel.COUNT_LIMIT:
            raise McsaCountsLimit(self.path, counts)
        return counts
